/*
 * Check and change the strategy of page out.
 * If almost the all file caches are missed, it may be better
 * to pageout the new page insteadof the old, or it will miss
 * again when read the file.
 *
 * NOTION:
 *    LRC： Least Recently Creation
 *    MRC:  Most Recently Creation.
 *
 * MRC algorithm: pageout new page (dequeue from inactive list's head).
 * LRC algorithm: pageout old page (dequeue from inactive list's tail).
 *
 *   Copyright @nfschina
 *   Author:liqiong(liqiong@nfschina.com)
 */

#ifndef _LINUX_FILEMAP_PAGEOUT_H
#define _LINUX_FILEMAP_PAGEOUT_H

#include <linux/mmzone.h>

#define lru_to_page_mrc(head) (list_entry((head)->next, struct page, lru))

#ifdef ARCH_HAS_PREFETCHW
#define prefetchw_next_lru_page(_page, _base, _field)			\
	do {								\
		if ((_page)->lru.next != _base) {			\
			struct page *next;				\
									\
			next = lru_to_page_mrc(&(_page->lru));		\
			prefetchw(&next->_field);			\
		}							\
	} while (0)
#else
#define prefetchw_next_lru_page(_page, _base, _field) do { } while (0)
#endif

enum filemap_stat_flags {
	FMAP_MRC,  /* reclaim the most recently used file map page */
};

extern int sysctl_filemap_hit_ratio;
extern int sysctl_filemap_mrc_thres;

/* The most time use mrc, when counter down to 0, chang to lrc */
extern int sysctl_filemap_mrc_duration;

/* After boot or chang pagout strategy, need time to check cache status,
 * which called window -- timing for check hit ratio again and decide the
 * next move.
 *
 * The window's unit is the max free pages (~memory size).
 *
 * Foramt of this 'int' data:
 *     - the first  byte is mrc  window
 *     - the second byte is lrc  window
 *     - the third  byte is boot window
 * eg: 0x00060307:
 *     - boot window : 0x06 -- 6/10 * unit
 *     - lrc window:   0x03 -- 3/10 * unit.
 *     - mrc window:   0x07 -- 7/10 * unit.
 */
extern int sysctl_filemap_mrc_win_factor;
extern int sysctl_filemap_mrc_wb_pages;
extern int sysctl_filemap_mrc_disable;
extern int filemap_mrc_win_factor;

extern void check_filemap_pageout_mrc(struct pglist_data *pgdat);
extern void filemap_pageout_mrc_writeback(struct page *page);

static inline void filemap_stat_inc(struct pglist_data *pgdat,
				unsigned int nr_pages, enum FILEMAP_STAT_ITEMS item)
{
	pgdat->fmap_stat[item] += nr_pages;
}

static inline void filemap_stat_clear(struct pglist_data *pgdat)
{
	enum FILEMAP_STAT_ITEMS i;
	for (i = NR_FMAP_HIT; i < NR_FMAP_STAT_ITEMS; i++)
		pgdat->fmap_stat[i] = 0;

	filemap_mrc_win_factor = sysctl_filemap_mrc_win_factor >> 16;
}

static inline void filemap_stat_reset_count(struct pglist_data *pgdat)
{
	enum FILEMAP_STAT_ITEMS i;
	for (i = NR_FMAP_HIT; i < NR_FMAP_FLAGS; i++)
		pgdat->fmap_stat[i] = 0;
}

static inline int filemap_hit_ratio(struct pglist_data *pgdat)
{
	unsigned long *stat = pgdat->fmap_stat;
	unsigned long total = stat[NR_FMAP_HIT] + stat[NR_FMAP_MISS];
	unsigned long hits = stat[NR_FMAP_HIT] > stat[NR_FMAP_RD] ?
						stat[NR_FMAP_HIT] - stat[NR_FMAP_RD] : 0;
	if (likely(total))
		return	hits * 100 / total;
	return 0;
}

static inline void filemap_use_mrc(struct pglist_data *pgdat)
{
	set_bit(FMAP_MRC, &pgdat->fmap_stat[NR_FMAP_FLAGS]);
}

static inline void filemap_use_lrc(struct pglist_data *pgdat)
{
	clear_bit(FMAP_MRC, &pgdat->fmap_stat[NR_FMAP_FLAGS]);
}

static inline int filemap_is_mrc_used(struct pglist_data *pgdat)
{
	return test_bit(FMAP_MRC, &pgdat->fmap_stat[NR_FMAP_FLAGS]);
}

#endif
